@@ -1,6 +1,6 @@
 package VCS::CMSynergy::Users;
 
-our $VERSION = sprintf("%d.%02d", q%version: 1.09 % =~ /(\d+)\.(\d+)/);
+our $VERSION = do { (my $v = q%version: 11 %) =~ s/.*://; sprintf("%d.%02d", split(/\./, $v), 0) };
 
 =head1 NAME
 
@@ -58,18 +58,37 @@ with the contents of C<$hashref>. Duplicate roles will be removed
 from C<$hashref>'s values before writing back the table.
 You must be in the B<ccm_admin> role to use this form. 
 
+All operations try to preserve the order of roles (L<add_roles> appends
+the roles that are actually new for the user). This mostly matters for
+the role listed first for a user, as CM Synergy uses this as default
+role for the user's session when the user calls B<ccm start> without
+the B<-r> option.
+
+If you have one of the modules L<Tie::Hash::Indexed> or L<Tie::IxHash> 
+installed, than the value returned by L</users> is actually 
+a reference to a tied hash that preserves the insertion order of keys.
+In this case, the order of hash keys reflects the order 
+of user lines in the I<users> "file".
+
 Note that
 
   $ccm->users($ccm->users);
 
-results in a functionally equivalent, though probably not identical
-users table as before (due to hash key ordering etc).
+always results in a functionally equivalent users table. The order
+of user lines may have changed, though, unless an order preserving hash
+was used as decribed above.
 
 Note: For typical CM Synergy administrator usage
 it is usually more convenient to use one of the methods below.
 
 =cut
 
+my $UseIndexedHash;
+foreach (qw/Tie::Hash::Indexed Tie::IxHash/)
+{
+    $UseIndexedHash = $_, last if eval "require $_";
+}
+
 sub users
 {
     my $self = shift;
@@ -81,15 +100,16 @@ sub users
     if (@_ == 0)
     {
 	my $text = $self->get_attribute(users => $self->base_model);
-	return undef unless defined $text;
+	return unless defined $text;
 	
 	my $users = {};
+	tie %$users, $UseIndexedHash if $UseIndexedHash;
+
 	foreach (split(/\n/, $text))
 	{
 	    my ($user, $roles) = /^ \s* user \s+ (\S+) \s* = \s* (.*) ;/x;
 	    next unless defined $user;
-	    my @roles =  split(" ", $roles);
-	    $users->{$user} = \@roles;
+	    $users->{$user} = [ split(" ", $roles) ];
 	}
 
 	return $users;
@@ -100,18 +120,16 @@ sub users
 	unless ref $users eq "HASH";
 
     my $text = "";
-    foreach my $user (sort keys %$users)
+    while (my ($user, $roles) = each %$users)
     {
-	my $roles = $users->{$user};
 	return $self->set_error("illegal value for user `$user' (array ref expected)")
 	    unless ref $roles eq "ARRAY";
-	return $self->set_error("no roles defined for user `$user")
+	return $self->set_error("no roles defined for user `$user'")
 	    unless @$roles;
 
 	# remove duplicates
-	my %roles = map { $_ => 1 } @$roles;
-
-	$text .= "user $user = " . join(" ", sort keys %roles) . ";\n";
+	my %dup;
+	$text .= "user $user = " . join(" ", grep { !$dup{$_}++ } @$roles) . ";\n";
     }
 
     my ($rc, $out, $err) = $self->ccm_with_text_editor($text, qw(users));
@@ -133,7 +151,7 @@ sub add_user
     my ($self, $user, @roles) = @_;
 
     my $users = $self->users;
-    return undef unless $users;
+    return unless $users;
 
     $users->{$user} = \@roles;
 
@@ -153,7 +171,7 @@ sub delete_user
     my ($self, $user) = @_;
 
     my $users = $self->users;
-    return undef unless $users;
+    return unless $users;
 
     delete $users->{$user};
 
@@ -174,7 +192,7 @@ sub add_roles
     my ($self, $user, @roles) = @_;
 
     my $users = $self->users;
-    return undef unless $users;
+    return unless $users;
 
     return $self->set_error("user `$user' doesn't exist")
 	unless $users->{$user};
@@ -198,13 +216,15 @@ sub delete_roles
     my ($self, $user, @roles) = @_;
 
     my $users = $self->users;
-    return undef unless $users;
+    return unless $users;
 
     return $self->set_error("user `$user' doesn't exist")
 	unless exists $users->{$user};
  
-    my %to_be_deleted = map { $_ => 1 } @roles;
-    @{ $users->{$user} } = grep { !$to_be_deleted{$_} } @{ $users->{$user} };
+    my %del;
+    @del{@roles} = ();
+
+    $users->{$user} = [ grep { !exists $del{$_} } @{ $users->{$user} } ];
 
     $self->users($users);
 }
@@ -223,7 +243,7 @@ sub get_roles
     my ($self, $user) = @_;
 
     my $users = $self->users;
-    return undef unless $users;
+    return unless $users;
 
     my $roles = $users->{$user};
     return $roles ? @$roles : ();
